"""Registry service commands."""

import enum
import typing
from uuid import UUID

import httpx
import typer

from neosctl import constant, util
from neosctl.schema import Common
from neosctl.services.registry import schema
from neosctl.util import process_response

app = typer.Typer()
product_app = typer.Typer()
core_app = typer.Typer()
mesh_app = typer.Typer()

app.add_typer(product_app, name="product", help="Manage products.")
app.add_typer(core_app, name="core", help="Manage cores.")
app.add_typer(mesh_app, name="mesh", help="Manage mesh.")

ACCOUNT_ARG = typer.Option(None, help="Account override (root only).", callback=util.sanitize)


class SubscriptionStatusEnum(enum.Enum):
    """Enum representing status of a subscription."""

    active = "active"
    inactive = "inactive"


def _core_url(obj: Common) -> str:
    return "{}/registry/core".format(obj.get_hub_api_url().rstrip("/"))


def _mesh_core_url(obj: Common) -> str:
    return "{}/registry/mesh/core".format(obj.get_hub_api_url().rstrip("/"))


def _mesh_subscriptions_url(obj: Common) -> str:
    return "{}/registry/mesh/subscriptions".format(obj.get_hub_api_url().rstrip("/"))


def _data_product_url(obj: Common, postfix: str = "") -> str:
    return "{}/registry/data_product{}".format(obj.get_hub_api_url().rstrip("/"), postfix)


def _mesh_data_product_url(obj: Common, identifier: str) -> str:
    return "{}/registry/mesh/core/{}/data_product".format(obj.get_hub_api_url().rstrip("/"), identifier)


def _subscribe_data_product_url(obj: Common, identifier: str) -> str:
    return "{}/registry/core/{}/data_product/subscribe".format(obj.get_hub_api_url().rstrip("/"), identifier)


def _subscription_data_product_url(obj: Common, identifier: str) -> str:
    return "{}/registry/core/{}/data_product/subscription".format(obj.get_hub_api_url().rstrip("/"), identifier)


@app.command(name="register-core", help="Register a core. [DEPRECATED]")
@core_app.command(name="register")
def register_core(
    ctx: typer.Context,
    partition: str = typer.Argument(..., help="Core partition", callback=util.sanitize),
    name: str = typer.Argument(..., help="Core name", callback=util.sanitize),
    account: typing.Optional[str] = ACCOUNT_ARG,
    *,
    private: bool = typer.Option(
        False,
        "--private",
        help="Limit visibility in mesh to core account.",
        callback=util.sanitize,
    ),
) -> None:
    """Register a core.

    Register a core to receive an identifier and access key for use in deployment.
    """

    @util.ensure_login
    def _request(ctx: typer.Context, rc: schema.RegisterCore) -> httpx.Response:
        return util.post(
            ctx,
            constant.REGISTRY,
            _core_url(ctx.obj),
            json=rc.model_dump(exclude_none=True),
            headers={"X-Partition": partition},
            account=account,
        )

    rc = schema.RegisterCore(name=name, public=not private)

    r = _request(ctx, rc)
    process_response(r)


@app.command(name="migrate-core", help="Migrate a core. [DEPRECATED]")
@core_app.command(name="migrate")
def migrate_core(
    ctx: typer.Context,
    identifier: str = typer.Option(..., help="Core identifier"),
    urn: str = typer.Option(..., help="Core urn", callback=util.sanitize),
    account: str = typer.Option(..., help="Account name", callback=util.sanitize),
) -> None:
    """Migrate a core out of root account.

    Migrate a core from `root` into an actual account.
    """

    @util.ensure_login
    def _request(ctx: typer.Context, mc: schema.MigrateCore) -> httpx.Response:
        return util.post(
            ctx,
            constant.REGISTRY,
            f"{_core_url(ctx.obj)}/{identifier}/migrate",
            json=mc.model_dump(),
        )

    mc = schema.MigrateCore(
        urn=urn,
        account=account,
    )

    r = _request(ctx, mc)
    process_response(r)


@app.command(name="list-cores", help="List accessible cores. [DEPRECATED]")
@core_app.command(name="list")
def list_cores(
    ctx: typer.Context,
    account: typing.Optional[str] = ACCOUNT_ARG,
) -> None:
    """List accessible cores."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.get(
            ctx,
            constant.REGISTRY,
            _core_url(ctx.obj),
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@app.command(name="remove-core", help="Remove a registered core. [DEPRECATED]")
@core_app.command(name="remove")
def remove_core(
    ctx: typer.Context,
    identifier: str = typer.Option(..., help="Core identifier"),
    account: typing.Optional[str] = ACCOUNT_ARG,
) -> None:
    """Remove a registered core."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.delete(
            ctx,
            constant.REGISTRY,
            f"{_core_url(ctx.obj)}/{identifier}",
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@app.command(name="search", help="Search published data products. [DEPRECATED]")
@product_app.command(name="search")
def search_products(
    ctx: typer.Context,
    search_term: str = typer.Argument(..., callback=util.sanitize),
    *,
    keyword: bool = typer.Option(False, "--keyword/--hybrid", help="Search mode"),
) -> None:
    """Search published data products across cores."""

    @util.ensure_login
    def _request(ctx: typer.Context, search_term: str, *, keyword: bool) -> httpx.Response:
        return util.get(
            ctx,
            constant.REGISTRY,
            _data_product_url(ctx.obj, "/search"),
            params={"search_term": search_term, "keyword_search": keyword},
        )

    r = _request(ctx, search_term, keyword=keyword)
    process_response(r)


@app.command(name="get-product", help="Get data product details. [DEPRECATED]")
@product_app.command(name="get")
def get_product(
    ctx: typer.Context,
    urn: str = typer.Argument(..., callback=util.sanitize),
) -> None:
    """Get data product details."""

    @util.ensure_login
    def _request(ctx: typer.Context, urn: str) -> httpx.Response:
        return util.get(
            ctx,
            constant.REGISTRY,
            _data_product_url(ctx.obj, f"/urn/{urn}"),
        )

    r = _request(ctx, urn)
    process_response(r)


@app.command(name="upsert-contact", help="Add/Update a contact for a core. [DEPRECATED]")
@core_app.command(name="upsert-contact")
def upsert_contact(
    ctx: typer.Context,
    identifier: str = typer.Option(..., help="Core identifier"),
    user_id: UUID = typer.Option(..., help="Contact id"),
    role: str = typer.Option(..., help="Contact role"),
    account: typing.Optional[str] = ACCOUNT_ARG,
) -> None:
    """Add/Update a contact for a core."""

    @util.ensure_login
    def _request(ctx: typer.Context, c: schema.AddCoreContact) -> httpx.Response:
        return util.post(
            ctx,
            constant.REGISTRY,
            f"{_core_url(ctx.obj)}/{identifier}/contact",
            json=c.model_dump(),
            account=account,
        )

    mc = schema.AddCoreContact(
        user_id=str(user_id),
        role=role,
    )

    r = _request(ctx, mc)
    process_response(r)


@app.command(name="remove-contact", help="Remove a contact for a core. [DEPRECATED]")
@core_app.command(name="remove-contact")
def remove_contact(
    ctx: typer.Context,
    identifier: str = typer.Option(..., help="Core identifier"),
    user_id: UUID = typer.Option(..., help="Contact id"),
    account: typing.Optional[str] = ACCOUNT_ARG,
) -> None:
    """Remove a contact for a core."""

    @util.ensure_login
    def _request(ctx: typer.Context, c: schema.RemoveCoreContact) -> httpx.Response:
        return util.delete(
            ctx,
            constant.REGISTRY,
            f"{_core_url(ctx.obj)}/{identifier}/contact",
            json=c.model_dump(),
            account=account,
        )

    mc = schema.RemoveCoreContact(
        user_id=str(user_id),
    )

    r = _request(ctx, mc)
    process_response(r)


@app.command(name="mesh-cores", help="List visible cores. [DEPRECATED]")
@mesh_app.command(name="cores")
def mesh_cores(
    ctx: typer.Context,
    account: typing.Optional[str] = ACCOUNT_ARG,
    search: str = typer.Option(None, help="Search core name(s)", callback=util.sanitize),
) -> None:
    """List visible cores."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.get(
            ctx,
            constant.REGISTRY,
            _mesh_core_url(ctx.obj),
            params={"search": search} if search else None,
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@app.command(name="mesh-core-products", help="List visible products in a core. [DEPRECATED]")
@mesh_app.command(name="core-products")
def mesh_core_products(
    ctx: typer.Context,
    identifier: str = typer.Option(..., help="Core identifier"),
    account: typing.Optional[str] = ACCOUNT_ARG,
) -> None:
    """List visible products in a core."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.get(
            ctx,
            constant.REGISTRY,
            _mesh_data_product_url(ctx.obj, identifier),
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@app.command(name="mesh-subscriptions", help="List mesh subscriptions. [DEPRECATED]")
@mesh_app.command(name="subscriptions")
def mesh_subscriptions(
    ctx: typer.Context,
    account: typing.Optional[str] = ACCOUNT_ARG,
) -> None:
    """List mesh subscriptions."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.get(
            ctx,
            constant.REGISTRY,
            _mesh_subscriptions_url(ctx.obj),
            account=account,
        )

    r = _request(ctx)
    process_response(r)


@app.command(name="subscribe-product", help="Subscribe to a data product. [DEPRECATED]")
@product_app.command(name="subscribe")
def subscribe_product(
    ctx: typer.Context,
    identifier: str = typer.Option(
        ...,
        "--core-id",
        "-cid",
        help="Core identifier",
    ),
    product_identifier: str = typer.Option(
        ...,
        "--product-id",
        "-pid",
        help="Data product identifier",
    ),
    account: typing.Optional[str] = ACCOUNT_ARG,
) -> None:
    """Subscribe to a data product."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.post(
            ctx,
            constant.REGISTRY,
            _subscribe_data_product_url(ctx.obj, identifier),
            account=account,
            json={"data_product_urn": product_identifier},
        )

    r = _request(ctx)
    process_response(r)


@app.command(name="unsubscribe-product", help="Unsubscribe from a data product. [DEPRECATED]")
@product_app.command(name="unsubscribe")
def unsubscribe_product(
    ctx: typer.Context,
    identifier: str = typer.Option(
        ...,
        "--core-id",
        "-cid",
        help="Core identifier",
    ),
    product_identifier: str = typer.Option(
        ...,
        "--product-id",
        "-pid",
        help="Data product identifier",
    ),
    account: typing.Optional[str] = ACCOUNT_ARG,
) -> None:
    """Unsubscribe from a data product."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.delete(
            ctx,
            constant.REGISTRY,
            _subscribe_data_product_url(ctx.obj, identifier),
            account=account,
            json={"data_product_urn": product_identifier},
        )

    r = _request(ctx)
    process_response(r)


@app.command(name="update-subscription", help="Update subscription to a data product. [DEPRECATED]")
@product_app.command(name="update-subscription")
def update_subscription(
    ctx: typer.Context,
    identifier: str = typer.Option(
        ...,
        "--core-id",
        "-cid",
        help="Core identifier",
    ),
    product_identifier: str = typer.Option(
        ...,
        "--product-id",
        "-pid",
        help="Data product identifier",
    ),
    subsriber_core_identifier: str = typer.Option(
        ...,
        "--sub-core-id",
        "-scid",
        help="Subscriber core identifier.",
    ),
    reason: str = typer.Option(...),
    status: SubscriptionStatusEnum = typer.Option(...),
    account: typing.Optional[str] = ACCOUNT_ARG,
) -> None:
    """Update subscription to a data product."""

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.put(
            ctx,
            constant.REGISTRY,
            _subscription_data_product_url(ctx.obj, identifier),
            account=account,
            json={
                "data_product_urn": product_identifier,
                "subscriber_core_identifier": subsriber_core_identifier,
                "status": status.value,
                "reason": reason,
            },
        )

    r = _request(ctx)
    process_response(r)



"""Constant values for neosctl."""

import pathlib

PROFILE_FILENAME = "profile"
ENV_FILENAME = "env"
CREDENTIAL_FILENAME = "credential"
LOG_FILENAME = "error.log"
ROOT_FILEPATH = pathlib.Path().home() / ".neosctl"
PROFILE_FILEPATH = ROOT_FILEPATH / PROFILE_FILENAME
ENV_FILEPATH = ROOT_FILEPATH / ENV_FILENAME
CREDENTIAL_FILEPATH = ROOT_FILEPATH / CREDENTIAL_FILENAME
LOG_FILEPATH = ROOT_FILEPATH / LOG_FILENAME


DEFAULT_PROFILE = "default"


SUCCESS_CODE = 200
REDIRECT_CODE = 300
BAD_REQUEST_CODE = 400
UNAUTHORISED_CODE = 401

GATEWAY = "gateway"
REGISTRY = "registry"
IAM = "iam"



"""Utility functions for neosctl."""

import asyncio
import configparser
import json
import logging
import pathlib
import re
import sys
import typing

import click
import httpx
import pydantic
import rtoml
import typer
from neos_common.authorization.signer import KeyPair
from neos_common.client import base
from neos_common.error import HttpException
from pygments import formatters, highlight, lexers

from neosctl import constant, schema

logger = logging.getLogger("neosctl.error")

if sys.version_info < (3, 10):  # pragma: no cover
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec  # pragma: no cover


def dumps_formatted_json(
    payload: typing.Union[dict, list],
    *,
    sort_keys: bool = True,
) -> str:
    """Dump formatted json.

    json.dump provided payload with indent 2 and (un)sorted keys.
    """
    return json.dumps(payload, indent=2, sort_keys=sort_keys)


def prettify_json(payload: typing.Union[dict, list]) -> str:
    """Dump formatted json with colour highlighting and sorted keys."""
    return highlight(dumps_formatted_json(payload), lexers.JsonLexer(), formatters.TerminalFormatter())


def prettify_json_no_sort(payload: typing.Union[dict, list]) -> str:
    """Dump formatted json with colour highlighting and no keys sort."""
    return highlight(dumps_formatted_json(payload, sort_keys=False), lexers.JsonLexer(), formatters.TerminalFormatter())


def is_success_response(response: httpx.Response) -> bool:
    """Check if a response is `successful`."""
    if constant.SUCCESS_CODE <= response.status_code < constant.REDIRECT_CODE:
        return True
    return False


def exit_with_output(msg: str, exit_code: int = 0) -> typer.Exit:
    """Render output to terminal and exit."""
    typer.echo(msg)

    return typer.Exit(exit_code)


def process_response(
    response: httpx.Response,
    render_callable: typing.Callable[[typing.Union[dict, list]], str] = prettify_json,
) -> None:
    """Process a server response, render the output and exit."""
    exit_code = 0
    try:
        data = response.json()
        if response.status_code >= constant.BAD_REQUEST_CODE:
            exit_code = 1
            message = prettify_json(data)
        else:
            message = render_callable(data)
    except Exception:
        logger.info(response.content)
        logger.exception("Failure to parse response.")
        exit_code = 1
        message = "Unable to parse response."

    raise exit_with_output(
        msg=message,
        exit_code=exit_code,
    )


def read_config_dotfile() -> configparser.ConfigParser:
    """Read in `.neosctl/profile` configuration file and parse."""
    c = configparser.ConfigParser()
    if constant.ROOT_FILEPATH.is_file():
        # Move legacy profiles into `~/.neosctl folder`
        legacy = pathlib.Path(str(constant.ROOT_FILEPATH))
        backup = pathlib.Path(f"{constant.ROOT_FILEPATH}.bak")
        legacy.rename(str(backup))
        constant.ROOT_FILEPATH.mkdir()
        backup.rename(str(constant.PROFILE_FILEPATH))
    constant.ROOT_FILEPATH.mkdir(exist_ok=True)
    c.read(constant.PROFILE_FILEPATH)
    return c


def read_env_dotfile() -> dict:
    """Read in `.neosctl/env` configuration file and parse."""
    if pathlib.Path(constant.ENV_FILEPATH).exists():
        return rtoml.load(constant.ENV_FILEPATH)

    return {}


def read_credential_dotfile() -> configparser.ConfigParser:
    """Read in `.neosctl/credential` configuration file and parse."""
    c = configparser.ConfigParser()
    c.read(constant.CREDENTIAL_FILEPATH)
    return c


def get_optional_user_profile_section(
    c: configparser.ConfigParser,
    profile_name: str,
) -> typing.Optional[configparser.SectionProxy]:
    """Get profile from neosctl configuration.

    Returns:
    -------
    Return configparser.SectionProxy if found, or None.
    """
    try:
        return c[profile_name]
    except KeyError:
        return None


def get_user_profile_section(c: configparser.ConfigParser, profile_name: str) -> configparser.SectionProxy:
    """Get profile from neosctl configuration.

    If profile is not found exit cli.

    Returns:
    -------
    configparser.SectionProxy for the requested profile.
    """
    try:
        return c[profile_name]
    except KeyError:
        pass

    raise exit_with_output(
        msg=f"Profile {profile_name} not found.",
        exit_code=1,
    )


def get_env_section(c: dict, name: str) -> schema.Env:
    """Get env from neosctl configuration.

    If env is not found exit cli.

    Returns:
    -------
    schema.Env for the requested environment.
    """
    try:
        return schema.Env(name=name, **c[name])
    except KeyError:
        pass

    raise exit_with_output(
        msg=f"Environment {name} not found.",
        exit_code=1,
    )


def get_user_credential(
    c: configparser.ConfigParser,
    profile_name: str,
    *,
    optional: bool = False,
) -> typing.Union[schema.Credential, None]:
    """Get profile from neosctl credentials.

    If profile is not found exit cli.

    Returns:
    -------
    schema.Credential for the requested profile.
    """
    try:
        return schema.Credential(**c[profile_name])
    except KeyError:
        pass

    if not optional:
        raise exit_with_output(
            msg=f"Profile {profile_name} not configured with credentials.",
            exit_code=1,
        )
    return None


def _auth_url(iam_api_url: str) -> str:
    return "{}".format(iam_api_url.rstrip("/"))


def check_refresh_token_exists(ctx: typer.Context) -> bool:
    """Check if refresh token exists."""
    if ctx.obj.active_env:
        if not ctx.obj.active_env.refresh_token:
            raise exit_with_output(
                msg="You need to login. Run neosctl env login",
                exit_code=1,
            )
        return True

    if not ctx.obj.profile.refresh_token:  # nosec: B105
        raise exit_with_output(
            msg=f"You need to login. Run neosctl -p {ctx.obj.profile_name} auth login",
            exit_code=1,
        )

    return True


def get_refresh_token(ctx: typer.Context) -> str:
    """Check if refresh token exists."""
    if ctx.obj.active_env:
        return ctx.obj.active_env.refresh_token

    return ctx.obj.profile.refresh_token  # nosec: B105


def _refresh_token(ctx: typer.Context) -> httpx.Response:
    check_profile_exists(ctx)
    check_refresh_token_exists(ctx)

    refresh_token = get_refresh_token(ctx)
    r = post(
        ctx,
        constant.IAM,
        f"{_auth_url(ctx.obj.get_hub_api_url())}/iam/refresh",
        json={"refresh_token": refresh_token},
    )

    if not is_success_response(r):
        process_response(r)

    if ctx.obj.active_env:
        d = r.json()
        ctx.obj.active_env.access_token = d["access_token"]
        ctx.obj.active_env.refresh_token = d["refresh_token"]

        upsert_env(ctx, ctx.obj.active_env)
    upsert_config(ctx, update_profile(ctx, schema.Auth(**r.json())))

    return r


P = ParamSpec("P")


def ensure_login(method: typing.Callable[P, httpx.Response]) -> typing.Callable[P, httpx.Response]:
    """Capture authentication errors and retry requests.

    On request failure check if the response is a 401, and attempt to refresh the access_token.
    Retry the request with refreshed token, on subsequent failure, return.
    """

    def check_access_token(*args: P.args, **kwargs: P.kwargs) -> httpx.Response:
        ctx = args[0]
        if not isinstance(ctx, (typer.Context, click.Context)):
            # Developer reminder
            msg = "First argument should be click.Context instance"
            raise TypeError(msg)

        r = method(*args, **kwargs)

        check_profile_exists(ctx)

        # Try to refresh token
        # Confirm it is a token invalid 401, registry not configured mistriggers this flow.
        if r.status_code == constant.UNAUTHORISED_CODE:
            data = r.json()
            if "type" in data and data["type"] in {
                "authorization-required",
                "authorization-failed",
                "invalid-authorization",
            }:
                _refresh_token(ctx)

                # Refresh the context
                c = read_config_dotfile()
                ctx.obj.config = c
                ctx.obj.profile = get_user_profile(c, ctx.obj.profile_name)

                r = method(*args, **kwargs)

        return r

    return check_access_token


def update_profile(
    ctx: typer.Context,
    auth: schema.Auth = schema.Auth(),
) -> schema.Profile:
    """Update user profile."""
    return schema.Profile(
        gateway_api_url=ctx.obj.gateway_api_url,
        hub_api_url=ctx.obj.hub_api_url,
        storage_api_url=ctx.obj.storage_api_url,
        user=ctx.obj.profile.user,
        access_token=auth.access_token,
        refresh_token=auth.refresh_token,
        ignore_tls=ctx.obj.profile.ignore_tls,
    )


def get_user_profile(
    c: configparser.ConfigParser,
    profile_name: str,
    *,
    allow_missing: bool = False,
) -> typing.Optional[typing.Union[schema.Profile, schema.OptionalProfile]]:
    """Get user profile.

    If allow_missing if False and the profile is not found, the cli will exit with an error output.
    """
    if allow_missing:
        profile_config = get_optional_user_profile_section(c, profile_name)
    else:
        profile_config = get_user_profile_section(c, profile_name)

    if profile_config:
        try:
            return schema.Profile(**profile_config)
        except pydantic.ValidationError as e:
            required_fields = [str(err["loc"][0]) for err in e.errors() if err["msg"] == "field required"]
            raise exit_with_output(  # noqa: TRY200, B904
                msg=("Profile dotfile doesn't include fields:\n  {}\nUse neosctl -p {} profile init").format(
                    "\n  ".join(required_fields),
                    profile_name,
                ),
                exit_code=1,
            )
    return None


def get_active_env(
    c: dict,
) -> typing.Optional[schema.Env]:
    """Get active environment."""
    for name, data in c.items():
        if data["active"]:
            return schema.Env(name=name, **data)

    return None


def bearer(ctx: typer.Context) -> typing.Optional[dict]:
    """Generate bearer authorization header."""
    if not (ctx.obj.profile and ctx.obj.profile.access_token):  # nosec: B105
        return None

    return {"Authorization": f"Bearer {ctx.obj.profile.access_token}"}


def check_profile_exists(ctx: typer.Context) -> bool:
    """Check if a profile (or env) exists in neosctl configuration or exit."""
    if ctx.obj.active_env:
        return True

    if not ctx.obj.profile:
        raise exit_with_output(
            msg=f"Profile not found! Run neosctl -p {ctx.obj.profile_name} profile init",
            exit_code=1,
        )

    return True


def check_env_active(ctx: typer.Context) -> bool:
    """Check if a profile (or env) exists in neosctl configuration or exit."""
    if not ctx.obj.active_env:
        raise exit_with_output(
            msg="Environment not active! Run neosctl env activate <env>",
            exit_code=1,
        )

    return True


def upsert_config(
    ctx: typer.Context,
    profile: schema.Profile,
) -> configparser.ConfigParser:
    """Update neosctl configuration profile in place."""
    ctx.obj.config[ctx.obj.profile_name] = profile.model_dump()

    with constant.PROFILE_FILEPATH.open("w") as profile_file:
        ctx.obj.config.write(profile_file)

    return ctx.obj.config


def upsert_env(
    ctx: typer.Context,
    env: schema.Env,
) -> configparser.ConfigParser:
    """Update neosctl env configuration in place."""
    ctx.obj.env[env.name] = env.model_dump(exclude=["name"], mode="json")

    with constant.ENV_FILEPATH.open("w") as env_file:
        rtoml.dump(ctx.obj.env, env_file)

    return ctx.obj.env


def upsert_credential(
    ctx: typer.Context,
    access_key_id: str,
    secret_access_key: str,
    name: typing.Optional[str] = None,
) -> None:
    """Update neosctl credential profile in place."""
    name = name or ctx.obj.profile_name
    ctx.obj.credential[name] = {"access_key_id": access_key_id, "secret_access_key": secret_access_key}

    with constant.CREDENTIAL_FILEPATH.open("w") as profile_file:
        ctx.obj.credential.write(profile_file)


def remove_config(
    ctx: typer.Context,
) -> configparser.ConfigParser:
    """Remove a profile from neosctl configuration."""
    if not ctx.obj.config.remove_section(ctx.obj.profile_name):
        raise exit_with_output(
            msg=f"Can not remove {ctx.obj.profile_name} profile, profile not found.",
            exit_code=1,
        )
    ctx.obj.credential.remove_section(ctx.obj.profile_name)

    with constant.PROFILE_FILEPATH.open("w") as profile_file:
        ctx.obj.config.write(profile_file)

    with constant.CREDENTIAL_FILEPATH.open("w") as credential_file:
        ctx.obj.credential.write(credential_file)

    return ctx.obj.config


def remove_env(
    ctx: typer.Context,
    name: str,
) -> configparser.ConfigParser:
    """Remove an environment from neosctl configuration."""
    if name not in ctx.obj.env:
        raise exit_with_output(
            msg=f"Can not remove {name} environment, environment not found.",
            exit_code=1,
        )
    ctx.obj.env.pop(name)
    ctx.obj.credential.remove_section(name)

    with constant.ENV_FILEPATH.open("w") as env_file:
        rtoml.dump(ctx.obj.env, env_file)

    with constant.CREDENTIAL_FILEPATH.open("w") as credential_file:
        ctx.obj.credential.write(credential_file)

    return ctx.obj.env


def get_file_location(filepath: str) -> pathlib.Path:
    """Get a Path for the provided filepath, exit if not found."""
    fp = pathlib.Path(filepath)
    if not fp.exists():
        raise exit_with_output(
            msg=f"Can not find file: {fp}",
            exit_code=1,
        )
    return fp


def load_json_file(fp: pathlib.Path, content_type: str) -> typing.Union[dict, list]:
    """Load contents of json file, exit if not found."""
    with fp.open() as f:
        try:
            data = json.load(f)
        except json.decoder.JSONDecodeError:
            logger.exception("Error loading json file.")
            raise exit_with_output(  # noqa: TRY200, B904
                msg=f"Invalid {content_type} file, must be json format.",
                exit_code=1,
            )
            return []  # never reached as raise exit_with_output, but it makes type checker happy

    return data


T = typing.TypeVar("T", bound=pydantic.BaseModel)


def load_object(t: type[T], filepath: str, file_description: str) -> T:
    """Trying to read and parse JSON file into the given data type."""
    fp = get_file_location(filepath)
    obj = load_json_file(fp, file_description)
    if not isinstance(obj, dict):
        raise exit_with_output(
            msg=f"Require a json file containing an object, `{type(obj)}` provided.",
            exit_code=1,
        )
    try:
        return t(**obj)
    except pydantic.ValidationError as e:
        logger.exception("Error loading json file.")
        raise exit_with_output(
            msg=str(e),
            exit_code=1,
        ) from e


class GenericHttpClient(base.NeosClient):
    """Generic HTTP client supporting Bearer and NEOS-HMAC-SHA256 auth."""

    def __init__(  # noqa: D107
        self,
        service: str,
        token: typing.Union[str, None],
        key_pair: typing.Union[KeyPair, None],
    ) -> None:
        # TODO(edgy): Replace with better error
        assert token is not None or key_pair is not None  # noqa: S101

        self._service = service
        self._token = token
        self._key_pair = key_pair

    @property
    def service_name(self) -> str:
        """Service name, used in signed scope."""
        return self._service

    @property
    def token(self) -> typing.Union[str, None]:
        """Bearer token, if provided."""
        return self._token

    @property
    def key_pair(self) -> typing.Union[KeyPair, None]:
        """Signing keypair, if provided."""
        return self._key_pair

    def request(
        self,
        url: str,
        method: base.Method,
        params: typing.Union[dict, None] = None,
        headers: typing.Union[dict, None] = None,
        json: typing.Union[dict, None] = None,
        **kwargs,  # noqa: ANN003
    ) -> httpx.Response:
        """Proxy for httpx requests."""
        return asyncio.run(self._request(url, method, params, headers, json, **kwargs))


def _request(ctx: typer.Context, method: str, service: str, url: str, **kwargs: ...) -> httpx.Response:
    name = ctx.obj.active_env.name if ctx.obj.active_env else ctx.obj.profile_name
    credential = get_user_credential(ctx.obj.credential, name, optional=True)
    bearer = None
    if credential is None:
        bearer = ctx.obj.active_env.access_token if ctx.obj.active_env else ctx.obj.profile.access_token

    c = GenericHttpClient(
        service=service,
        token=bearer,
        key_pair=KeyPair(credential.access_key_id, credential.secret_access_key, "ksa") if credential else None,
    )

    headers = kwargs.pop("headers", None) or {}
    headers_, params_ = extract_account(ctx, account=kwargs.pop("account", None))
    headers_["X-Partition"] = "ksa"
    headers.update(headers_)

    params = kwargs.pop("params", None) or {}
    params.update(params_ or {})

    try:
        return c.request(
            url,
            base.Method[method],
            verify=not ctx.obj.profile.ignore_tls,
            params=params,
            headers=headers,
            **kwargs,
        )
    except HttpException as e:
        logger.exception("Request failed.")
        raise exit_with_output(
            msg=f"{service.title()} {method} request to {url} failed ({e.status}). [{e.type}]",
            exit_code=1,
        ) from e


def _request_and_process(ctx: typer.Context, method: str, service: str, url: str, **kwargs: ...) -> None:
    render_callable = kwargs.pop("render_callable", prettify_json)

    @ensure_login
    def internal_request(context: typer.Context) -> httpx.Response:
        return _request(context, method, service, url, **kwargs)

    response = internal_request(ctx)

    process_response(response, render_callable)


def get_and_process(ctx: typer.Context, service: str, url: str, **kwargs: ...) -> None:
    """Execute and process GET request."""
    _request_and_process(ctx, "GET", service, url, **kwargs)


def post_and_process(ctx: typer.Context, service: str, url: str, **kwargs: ...) -> None:
    """Execute and process POST request."""
    _request_and_process(ctx, "POST", service, url, **kwargs)


def put_and_process(ctx: typer.Context, service: str, url: str, **kwargs: ...) -> None:
    """Execute and process PUT request."""
    _request_and_process(ctx, "PUT", service, url, **kwargs)


def delete_and_process(ctx: typer.Context, service: str, url: str, **kwargs: ...) -> None:
    """Execute and process DELETE request."""
    _request_and_process(ctx, "DELETE", service, url, **kwargs)


def patch_and_process(ctx: typer.Context, service: str, url: str, **kwargs: ...) -> None:
    """Execute and process PATCH request."""
    _request_and_process(ctx, "PATCH", service, url, **kwargs)


def get(ctx: typer.Context, service: str, url: str, **kwargs: ...) -> httpx.Response:
    """Execute a GET request."""
    return _request(ctx, "GET", service, url, **kwargs)


def post(ctx: typer.Context, service: str, url: str, **kwargs: ...) -> httpx.Response:
    """Execute a POST request."""
    return _request(ctx, "POST", service, url, **kwargs)


def put(ctx: typer.Context, service: str, url: str, **kwargs: ...) -> httpx.Response:
    """Execute a PUT request."""
    return _request(ctx, "PUT", service, url, **kwargs)


def delete(ctx: typer.Context, service: str, url: str, **kwargs: ...) -> httpx.Response:
    """Execute a DELETE request."""
    return _request(ctx, "DELETE", service, url, **kwargs)


def sanitize(
    ctx: typer.Context,  # noqa: ARG001
    param: click.Parameter,  # noqa: ARG001
    value: typing.Optional[str],
) -> typing.Optional[str]:
    """Parameter's sanitize callback."""
    if value and isinstance(value, str):
        return value.rstrip("\r\n")

    return value


def validate_string_not_empty(
    ctx: typer.Context,
    param: click.Parameter,
    value: str,
) -> str:
    """String validation callback."""
    value = value.strip()

    if not value:
        message = "Value must be a non-empty string."
        raise typer.BadParameter(message, ctx=ctx, param=param)

    return value


def validate_strings_are_not_empty(
    ctx: typer.Context,
    param: click.Parameter,
    values: list[str],
) -> list[str]:
    """List of strings validation callback."""
    return [validate_string_not_empty(ctx, param, value) for value in values]


def validate_regex(
    pattern: str,
) -> typing.Callable[[typer.Context, click.Parameter, str], str]:
    """Regex validation callback."""

    def factory(
        ctx: typer.Context,
        param: click.Parameter,
        value: str,
    ) -> str:
        value = value.strip()

        if not re.match(pattern, value):
            message = f"Value does not satisfy the rule {pattern}"
            raise typer.BadParameter(message, ctx=ctx, param=param)

        return value

    return factory


def user_profile_callback(
    ctx: typer.Context,
) -> None:
    """Inject required user_profile into context."""
    user_profile = get_user_profile(ctx.obj.config, ctx.obj.profile_name)
    ctx.obj.profile = user_profile


def extract_account(
    ctx: typer.Context,
    account: typing.Optional[str] = None,
) -> tuple[dict, typing.Optional[dict]]:
    """Set up account header and params."""
    profile_account = ctx.obj.get_account()
    if profile_account != "root" and account is not None:
        raise exit_with_output(
            msg="Only root account admins can impersonate other accounts.",
            exit_code=1,
        )

    headers, params = {"X-Account": profile_account}, None
    if account:
        headers["X-Account-Override"] = account
        params = {"account": account}

    return headers, params



"""Gateway data unit commands."""

from typing import Optional

import typer

from neosctl import constant, util
from neosctl.services.gateway import entity, schema

app = typer.Typer(name="data_unit")


def _data_unit_url(ctx: typer.Context) -> str:
    return "{}/v2/data_unit".format(ctx.obj.get_gateway_api_url().rstrip("/"))


def _identified_data_unit_url(ctx: typer.Context, identifier: str) -> str:
    return "{}/v2/data_unit/{}".format(
        ctx.obj.get_gateway_api_url().rstrip("/"),
        identifier,
    )


arg_generator = entity.EntityArgGenerator("Data Unit")


@app.command(name="create")
def create_entity(
    ctx: typer.Context,
    label: str = arg_generator.label,
    name: str = arg_generator.name,
    description: str = arg_generator.description,
    owner: Optional[str] = arg_generator.owner_optional,
    contacts: list[str] = arg_generator.contacts,
    links: list[str] = arg_generator.links,
) -> None:
    """Create data unit."""
    entity.create_entity(
        ctx=ctx,
        url=_data_unit_url(ctx=ctx),
        label=label,
        name=name,
        description=description,
        owner=owner,
        contacts=contacts,
        links=links,
    )


@app.command(name="list")
def list_entities(ctx: typer.Context) -> None:
    """List data units."""
    entity.list_entities(ctx=ctx, url=_data_unit_url(ctx=ctx))


@app.command(name="get")
def get_entity(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
) -> None:
    """Get data unit."""
    entity.get_entity(
        ctx=ctx,
        url=_data_unit_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="delete")
def delete_entity(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
) -> None:
    """Delete data unit."""
    entity.delete_entity(
        ctx=ctx,
        url=_data_unit_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="get-info")
def get_entity_info(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
) -> None:
    """Get data unit info."""
    entity.get_entity_info(
        ctx=ctx,
        url=_data_unit_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="update")
def update_entity(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    label: str = arg_generator.label,
    name: str = arg_generator.name,
    description: str = arg_generator.description,
) -> None:
    """Update data unit."""
    entity.update_entity(
        ctx=ctx,
        url=_data_unit_url(ctx=ctx),
        identifier=identifier,
        label=label,
        name=name,
        description=description,
    )


@app.command(name="update-info")
def update_entity_info(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    owner: str = arg_generator.owner,
    contacts: list[str] = arg_generator.contacts,
    links: list[str] = arg_generator.links,
) -> None:
    """Update data unit info."""
    entity.update_entity_info(
        ctx=ctx,
        url=_data_unit_url(ctx=ctx),
        identifier=identifier,
        owner=owner,
        contacts=contacts,
        links=links,
    )


@app.command(name="get-journal")
def get_entity_journal(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    page: int = typer.Option(1, "--page", "-p", help="Page number", callback=util.sanitize),
    per_page: int = typer.Option(25, "--per-page", "-pp", help="Number of items per page", callback=util.sanitize),
) -> None:
    """Get data unit journal."""
    entity.get_entity_journal(
        ctx=ctx,
        url=_data_unit_url(ctx=ctx),
        identifier=identifier,
        page=page,
        per_page=per_page,
    )


@app.command(name="update-journal")
def update_entity_journal(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to journal note payload", callback=util.sanitize),
) -> None:
    """Update data unit journal."""
    entity.update_entity_journal(
        ctx=ctx,
        url=_data_unit_url(ctx=ctx),
        identifier=identifier,
        filepath=filepath,
    )


@app.command(name="get-links")
def get_entity_links(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
) -> None:
    """Get data unit links."""
    entity.get_entity_links(
        ctx=ctx,
        url=_data_unit_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="update-metadata")
def update_entity_metadata(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to metadata description", callback=util.sanitize),
) -> None:
    """Update data unit metadata."""
    data = util.load_object(schema.UpdateEntityMetadataRequest, filepath, "metadata")

    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_unit_url(ctx, identifier)}/metadata",
        json=data.model_dump(by_alias=True),
    )


@app.command(name="delete-metadata")
def delete_entity_metadata(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to metadata description", callback=util.sanitize),
) -> None:
    """Delete data unit metadata."""
    data = util.load_object(schema.DeleteEntityMetadataRequest, filepath, "metadata")

    util.delete_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_unit_url(ctx, identifier)}/metadata",
        json=data.model_dump(by_alias=True),
    )


@app.command(name="get-schema")
def get_entity_schema(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
) -> None:
    """Get data unit schema."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_unit_url(ctx, identifier)}/schema",
    )


@app.command(name="get-config")
def get_entity_config(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
) -> None:
    """Get data unit config."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_unit_url(ctx, identifier)}/config",
    )


@app.command(name="update-config")
def update_entity_config(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to config description", callback=util.sanitize),
) -> None:
    """Update data unit config."""
    data = util.load_object(schema.UpdateDataUnitConfiguration, filepath, "configuration")
    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_unit_url(ctx, identifier)}/config",
        json=data.model_dump(by_alias=True),
    )



import typer

from neosctl import constant
from neosctl.services.gateway import (
    data_product,
    data_source,
    data_system,
    data_unit,
    journal_note,
    link,
    output,
    secret,
    spark,
    tag,
)
from neosctl.util import user_profile_callback

app = typer.Typer(name=constant.GATEWAY, callback=user_profile_callback)
app.add_typer(data_system.app, name="data-system", help="Manage data system entity.")
app.add_typer(data_source.app, name="data-source", help="Manage data source entity.")
app.add_typer(data_unit.app, name="data-unit", help="Manage data unit entity.")
app.add_typer(data_product.app, name="data-product", help="Manage data product entity.")
app.add_typer(output.app, name="output", help="Manage output entity.")
app.add_typer(link.app, name="link", help="Manage links.")
app.add_typer(secret.app, name="secret", help="Manage secrets.")
app.add_typer(tag.app, name="tag", help="Manage tags.")
app.add_typer(spark.app, name="spark", help="Manage spark job.")
app.add_typer(journal_note.app, name="journal-note", help="Manage journal note element.")



"""Utilities for user authentication."""

import typing
from warnings import warn

import typer

from neosctl import schema, util
from neosctl.util import (
    check_profile_exists,
    exit_with_output,
    is_success_response,
    process_response,
    upsert_config,
)

app = typer.Typer()


def _auth_url(iam_api_url: str) -> str:
    return "{}".format(iam_api_url.rstrip("/"))


@app.command()
def login(
    ctx: typer.Context,
    password: typing.Optional[str] = typer.Option(None, "--password", "-p", callback=util.sanitize),
) -> None:
    """Login to neos."""
    warn(
        "neosctl auth is deprecated, please use neosctl env",
        FutureWarning,
        stacklevel=2,
    )
    check_profile_exists(ctx)

    if password is None:
        password = typer.prompt(
            f"[{ctx.obj.profile_name}] Enter password for user ({ctx.obj.profile.user})",
            hide_input=True,
        )

    r = util.post(
        ctx,
        "iam",
        f"{_auth_url(ctx.obj.get_hub_api_url())}/iam/login",
        json={"user": ctx.obj.profile.user, "password": password},
    )

    if not is_success_response(r):
        process_response(r)

    upsert_config(ctx, util.update_profile(ctx, schema.Auth(**r.json())))

    raise exit_with_output(
        msg="Login success",
        exit_code=0,
    )


@app.command()
def logout(ctx: typer.Context) -> None:
    """Logout from neos."""
    warn(
        "neosctl auth is deprecated, please use neosctl env",
        FutureWarning,
        stacklevel=2,
    )
    check_profile_exists(ctx)

    util.check_refresh_token_exists(ctx)

    r = util.post(
        ctx,
        "iam",
        f"{_auth_url(ctx.obj.get_hub_api_url())}/iam/logout",
        json={"refresh_token": ctx.obj.profile.refresh_token},
    )

    if not is_success_response(r):
        process_response(r)

    upsert_config(ctx, util.update_profile(ctx, schema.Auth()))

    raise exit_with_output(
        msg="Logout success",
        exit_code=0,
    )



"""Environment commands, like env initialisation, deletion, and listing."""

import re
import typing

import click
import httpx
import tabulate
import typer

from neosctl import constant, schema, util

app = typer.Typer()


r = re.compile(r"http[s]?:\/\/.*")


def _validate_url(value: str) -> str:
    """Validate a given url is a valid url with schema."""
    m = r.match(value)
    if m is None:
        msg = f"Invalid url, must match pattern: `{r}`."
        raise click.UsageError(msg)
    return value


@app.command()
def init(
    ctx: typer.Context,
    name: str,
    *,
    hub_api_url: str = typer.Option(..., "--hub-api-url", "-h", callback=util.sanitize),
    username: str = typer.Option(..., "--username", "-u", callback=util.sanitize),
    account: str = typer.Option(..., "--account", "-a", callback=util.sanitize),
    ignore_tls: bool = typer.Option(
        False,
        "--ignore-tls",
        help="Ignore TLS errors (useful in local/development environments",
    ),
) -> None:
    """Initialise an environment.

    Create an environment that can be reused in later commands to define which
    services to interact with, and which user to interact as.

    Call `init` on an existing environment will update it.
    """
    typer.echo(f"Initialising [{name}] environment.")

    existing = ctx.obj.env.get(name, {})
    env = schema.Env(  # nosec: B106
        name=name,
        user=username,
        access_token="",
        refresh_token="",
        ignore_tls=ignore_tls,
        account=account,
        active=existing.get("active", False),
        active_core=existing.get("active_core"),
        hub_api_url=_validate_url(hub_api_url),
    )

    util.upsert_env(ctx, env)


@app.command()
def activate(
    ctx: typer.Context,
    name: str,
) -> None:
    """Activate an environment.

    Activate an environment to use its configuration for subsequent requests.
    """
    typer.echo(f"Activating [{name}] environment.")

    for name_ in ctx.obj.env:
        env = util.get_env_section(ctx.obj.env, name_)
        env.active = name == name_

        util.upsert_env(ctx, env)


@app.command(name="list")
def list_envs(
    ctx: typer.Context,
) -> None:
    """List available environments."""
    data = []
    for name in ctx.obj.env:
        env = util.get_env_section(ctx.obj.env, name)
        data.append([name, "*" if env.active else ""])
    typer.echo(tabulate.tabulate(data, headers=("Environment Name", "Active")))


@app.command()
def delete(
    ctx: typer.Context,
    name: str,
) -> None:
    """Delete an environment."""
    typer.confirm(f"Remove [{name}] environment", abort=True)
    util.remove_env(ctx, name)


@app.command()
def view(
    ctx: typer.Context,
    name: str,
) -> None:
    """View configuration for an environment."""
    env = util.get_env_section(ctx.obj.env, name)
    raise util.exit_with_output(
        msg=util.prettify_json(env.model_dump(mode="json")),
    )


@app.command()
def active(
    ctx: typer.Context,
) -> None:
    """View configuration for active environment."""
    util.check_env_active(ctx)

    raise util.exit_with_output(
        msg=util.prettify_json(ctx.obj.active_env.model_dump(mode="json")),
    )


@app.command()
def credentials(
    ctx: typer.Context,
    name: str,
    access_key_id: str,
    secret_access_key: str,
) -> None:
    """Configure access keys for an environment."""
    util.upsert_credential(ctx, access_key_id, secret_access_key, name=name)


def _auth_url(iam_api_url: str) -> str:
    return "{}".format(iam_api_url.rstrip("/"))


@app.command()
def login(
    ctx: typer.Context,
    password: typing.Optional[str] = typer.Option(None, "--password", "-p", callback=util.sanitize),
) -> None:
    """Login to environment."""
    util.check_env_active(ctx)

    if password is None:
        password = typer.prompt(
            f"[{ctx.obj.active_env.name}] Enter password for user ({ctx.obj.active_env.user})",
            hide_input=True,
        )

    r = util.post(
        ctx,
        "iam",
        f"{_auth_url(ctx.obj.get_hub_api_url())}/iam/login",
        json={"user": ctx.obj.active_env.user, "password": password},
    )

    if not util.is_success_response(r):
        util.process_response(r)

    d = r.json()
    ctx.obj.active_env.access_token = d["access_token"]
    ctx.obj.active_env.refresh_token = d["refresh_token"]

    util.upsert_env(ctx, ctx.obj.active_env)

    raise util.exit_with_output(
        msg="Login success",
        exit_code=0,
    )


@app.command()
def logout(ctx: typer.Context) -> None:
    """Logout from neos."""
    util.check_env_active(ctx)

    util.check_refresh_token_exists(ctx)

    r = util.post(
        ctx,
        "iam",
        f"{_auth_url(ctx.obj.get_hub_api_url())}/iam/logout",
        json={"refresh_token": ctx.obj.active_env.refresh_token},
    )

    if not util.is_success_response(r):
        util.process_response(r)

    ctx.obj.active_env.access_token = ""
    ctx.obj.active_env.refresh_token = ""

    util.upsert_env(ctx, ctx.obj.active_env)

    raise util.exit_with_output(
        msg="Logout success",
        exit_code=0,
    )


def _core_url(hub_api_url: str) -> str:
    return "{}/registry/core".format(hub_api_url.rstrip("/"))


@app.command(name="list-cores")
def list_cores(
    ctx: typer.Context,
) -> None:
    """List available cores."""
    util.check_env_active(ctx)

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.get(
            ctx,
            constant.REGISTRY,
            _core_url(ctx.obj.get_hub_api_url()),
        )

    def tabulate_cores(data: dict[str, list]) -> str:
        ret = [
            [
                core["name"],
                core["version"],
                core["host"],
                core["account"],
                core["urn"],
                "*" if ctx.obj.active_env.active_core and core["name"] == ctx.obj.active_env.active_core.name else "",
            ]
            for core in data["cores"]
        ]
        return tabulate.tabulate(ret, headers=("Core Name", "Version", "Host", "Account", "Urn", "Active"))

    r = _request(ctx)
    util.process_response(r, render_callable=tabulate_cores)


@app.command(name="activate-core")
def activate_core(
    ctx: typer.Context,
    name: str,
) -> None:
    """Activate a core in current environment."""
    util.check_env_active(ctx)

    @util.ensure_login
    def _request(ctx: typer.Context) -> httpx.Response:
        return util.get(
            ctx,
            constant.REGISTRY,
            _core_url(ctx.obj.get_hub_api_url()),
        )

    r = _request(ctx)
    if not util.is_success_response(r):
        util.process_response(r)

    active_core = None
    data = r.json()
    for core in data["cores"]:
        if core["name"] == name:
            active_core = core

    if active_core is None:
        raise util.exit_with_output(
            msg=f"Core {name} not found.",
            exit_code=1,
        )

    ctx.obj.active_env.active_core = schema.Core(**active_core)

    util.upsert_env(ctx, ctx.obj.active_env)


@app.command(name="set-account")
def set_account(
    ctx: typer.Context,
    account: str,
) -> None:
    """Switch active environment account."""
    util.check_env_active(ctx)

    ctx.obj.active_env.account = account

    util.upsert_env(ctx, ctx.obj.active_env)



"""Gateway spark commands."""

from typing import Optional

import typer

from neosctl import constant, util

app = typer.Typer(name="output")


def _identified_spark_url(ctx: typer.Context, identifier: str) -> str:
    return "{}/v2/spark/{}".format(
        ctx.obj.get_gateway_api_url().rstrip("/"),
        identifier,
    )


IDENTIFIER_ARGUMENT = typer.Argument(..., help="Spark identifier", callback=util.sanitize)


@app.command(name="status")
def element_status(
    ctx: typer.Context,
    identifier: str = IDENTIFIER_ARGUMENT,
    suffix: Optional[str] = typer.Option("latest", "--suffix", "-s", help="Job run suffix"),
    run: Optional[str] = typer.Option(None, "--run", "-r", help="Job run"),
) -> None:
    """Get spark status."""
    params = {"suffix": suffix}
    if run is not None:
        params["run"] = run
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_spark_url(ctx, identifier)}",
        params=params,
    )


@app.command(name="log")
def element_log(
    ctx: typer.Context,
    identifier: str = IDENTIFIER_ARGUMENT,
    suffix: Optional[str] = typer.Option("latest", "--suffix", "-s", help="Job run suffix"),
    run: Optional[str] = typer.Option(None, "--run", "-r", help="Job run"),
) -> None:
    """Get spark logs."""
    params = {"suffix": suffix}
    if run is not None:
        params["run"] = run
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_spark_url(ctx, identifier)}/log",
        params=params,
    )


@app.command(name="history")
def element_history(
    ctx: typer.Context,
    identifier: str = IDENTIFIER_ARGUMENT,
    suffix: Optional[str] = typer.Option(None, "--suffix", "-s", help="Job run suffix"),
) -> None:
    """Get spark history."""
    if suffix:
        util.get_and_process(
            ctx,
            constant.GATEWAY,
            f"{_identified_spark_url(ctx, identifier)}/history/{suffix}",
        )
    else:
        util.get_and_process(
            ctx,
            constant.GATEWAY,
            f"{_identified_spark_url(ctx, identifier)}/history",
        )



"""Gateway tag commands."""

from typing import Optional

import typer

from neosctl import constant, util
from neosctl.services.gateway import schema

app = typer.Typer(name="output")


def _tag_url(ctx: typer.Context) -> str:
    return "{}/v2/tag".format(ctx.obj.get_gateway_api_url().rstrip("/"))


NAME_ARGUMENT = typer.Argument(..., help="Tag name")
SCOPE_ARGUMENT = typer.Argument(..., help="Tag scope")


@app.command(name="list")
def list_elements(
    ctx: typer.Context,
    scope: schema.TagScope = SCOPE_ARGUMENT,
    *,
    system_defined: bool = typer.Option(False, "--system-defined", "-s", help="System defined"),
    tag_filter: Optional[str] = typer.Option(None, "--filter", "-f", help="Filter query"),
) -> None:
    """List tags."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        _tag_url(ctx),
        params={
            "scope": scope.value,
            "system_defined": system_defined,
            "tag_filter": tag_filter,
        },
    )


@app.command(name="create")
def create_element(
    ctx: typer.Context,
    name: str = NAME_ARGUMENT,
    scope: schema.TagScope = SCOPE_ARGUMENT,
) -> None:
    """Create new tag."""
    data = schema.UpdateTag(
        tag=name,
        scope=scope.value,
    )

    util.post_and_process(
        ctx,
        constant.GATEWAY,
        _tag_url(ctx),
        json=data.model_dump(by_alias=True),
    )


@app.command(name="delete")
def delete_element(
    ctx: typer.Context,
    name: str = NAME_ARGUMENT,
    scope: schema.TagScope = SCOPE_ARGUMENT,
) -> None:
    """Delete tag."""
    data = schema.UpdateTag(
        tag=name,
        scope=scope.value,
    )
    util.delete_and_process(
        ctx,
        constant.GATEWAY,
        _tag_url(ctx),
        json=data.model_dump(by_alias=True),
    )



"""Gateway data source commands."""

from typing import Optional

import typer

from neosctl import constant, util
from neosctl.services.gateway import entity, schema

app = typer.Typer(name="data_source")


def _data_source_url(ctx: typer.Context) -> str:
    return "{}/v2/data_source".format(ctx.obj.get_gateway_api_url().rstrip("/"))


def _identified_data_source_url(ctx: typer.Context, identifier: str) -> str:
    return "{}/v2/data_source/{}".format(
        ctx.obj.get_gateway_api_url().rstrip("/"),
        identifier,
    )


arg_generator = entity.EntityArgGenerator("Data Source")


@app.command(name="create")
def create_entity(
    ctx: typer.Context,
    label: str = arg_generator.label,
    name: str = arg_generator.name,
    description: str = arg_generator.description,
    owner: Optional[str] = arg_generator.owner_optional,
    contacts: list[str] = arg_generator.contacts,
    links: list[str] = arg_generator.links,
) -> None:
    """Create data source."""
    entity.create_entity(
        ctx=ctx,
        url=_data_source_url(ctx=ctx),
        label=label,
        name=name,
        description=description,
        owner=owner,
        contacts=contacts,
        links=links,
    )


@app.command(name="list")
def list_entities(ctx: typer.Context) -> None:
    """List data sources."""
    entity.list_entities(ctx=ctx, url=_data_source_url(ctx=ctx))


@app.command(name="get")
def get_entity(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
) -> None:
    """Get data source."""
    entity.get_entity(
        ctx=ctx,
        url=_data_source_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="delete")
def delete_entity(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
) -> None:
    """Delete data source."""
    entity.delete_entity(
        ctx=ctx,
        url=_data_source_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="get-info")
def get_entity_info(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
) -> None:
    """Get data source info."""
    entity.get_entity_info(
        ctx=ctx,
        url=_data_source_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="update")
def update_entity(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    label: str = arg_generator.label,
    name: str = arg_generator.name,
    description: str = arg_generator.description,
) -> None:
    """Update data source."""
    entity.update_entity(
        ctx=ctx,
        url=_data_source_url(ctx=ctx),
        identifier=identifier,
        label=label,
        name=name,
        description=description,
    )


@app.command(name="update-info")
def update_entity_info(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    owner: str = arg_generator.owner,
    contacts: list[str] = arg_generator.contacts,
    links: list[str] = arg_generator.links,
) -> None:
    """Update data source info."""
    entity.update_entity_info(
        ctx=ctx,
        url=_data_source_url(ctx=ctx),
        identifier=identifier,
        owner=owner,
        contacts=contacts,
        links=links,
    )


@app.command(name="get-journal")
def get_entity_journal(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    page: int = typer.Option(1, "--page", "-p", help="Page number", callback=util.sanitize),
    per_page: int = typer.Option(25, "--per-page", "-pp", help="Number of items per page", callback=util.sanitize),
) -> None:
    """Get data source journal."""
    entity.get_entity_journal(
        ctx=ctx,
        url=_data_source_url(ctx=ctx),
        identifier=identifier,
        page=page,
        per_page=per_page,
    )


@app.command(name="update-journal")
def update_entity_journal(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to journal note payload", callback=util.sanitize),
) -> None:
    """Update data source journal."""
    entity.update_entity_journal(
        ctx=ctx,
        url=_data_source_url(ctx=ctx),
        identifier=identifier,
        filepath=filepath,
    )


@app.command(name="get-links")
def get_entity_links(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
) -> None:
    """Get data source links."""
    entity.get_entity_links(
        ctx=ctx,
        url=_data_source_url(ctx=ctx),
        identifier=identifier,
    )


@app.command(name="get-connection")
def get_entity_connection(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
) -> None:
    """Get data source connection."""
    util.get_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_source_url(ctx, identifier)}/connection",
    )


@app.command(name="update-connection")
def update_entity_connection(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to connection description", callback=util.sanitize),
) -> None:
    """Update data source connection."""
    data = util.load_object(schema.UpdateDataSourceConnection, filepath, "configuration")
    util.put_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_source_url(ctx, identifier)}/connection",
        json=data.model_dump(by_alias=True),
    )


@app.command(name="set-connection-secrets")
def set_entity_connection_secrets(
    ctx: typer.Context,
    identifier: str = arg_generator.identifier,
    filepath: str = typer.Argument(..., help="Filepath to secrets description", callback=util.sanitize),
) -> None:
    """Set data source connection secrets."""
    data = util.load_object(schema.UpdateDataSourceConnectionSecret, filepath, "secret")
    util.post_and_process(
        ctx,
        constant.GATEWAY,
        f"{_identified_data_source_url(ctx, identifier)}/secret",
        json=data.secrets,
    )



"""IAM service schemas."""

from enum import Enum
from typing import Optional
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field


class EffectEnum(Enum):
    allow = "allow"
    deny = "deny"


class Statement(BaseModel):
    model_config = ConfigDict(use_enum_values=True)
    sid: str
    principal: UUID
    action: list[str]
    resource: list[str]
    condition: Optional[list[str]] = None
    effect: EffectEnum = EffectEnum.allow


class Statements(BaseModel):
    statements: list[Statement]


class Policy(BaseModel):
    version: str = "2022-10-01"
    statements: list[Statement]


class UserPolicy(BaseModel):
    user: UUID
    policy: Policy


class CreateAccount(BaseModel):
    name: str = Field(..., pattern=r"^[a-z][a-z0-9_-]*$", max_length=50, min_length=3)
    display_name: str
    description: str
    owner: str


class UpdateAccount(BaseModel):
    display_name: str
    description: str
    owner: str


GROUP_NAME_FIELD = Field(pattern=r"^[a-zA-Z0-9_][a-zA-Z0-9_\- ]{0,254}$")
GROUP_DESCRIPTION_FIELD = Field(None, max_length=1000)


class CreateUpdateGroup(BaseModel):
    name: str = GROUP_NAME_FIELD
    description: Optional[str] = GROUP_DESCRIPTION_FIELD


class Principals(BaseModel):
    principals: list[str]


class CreateUser(BaseModel):
    username: str
    enabled: bool
    first_name: str
    last_name: str
    email: str



"""Gateway entities link commands."""

import enum

import typer

from neosctl import constant, util
from neosctl.util import exit_with_output

app = typer.Typer()


def _link_url(
    ctx: typer.Context,
    p_type: str,
    p_identifier: str,
    c_type: str,
    c_identifier: str,
) -> str:
    return "{}/v2/link/{}/{}/{}/{}".format(
        ctx.obj.get_gateway_api_url().rstrip("/"),
        p_type,
        p_identifier,
        c_type,
        c_identifier,
    )


class ParentEntityType(enum.Enum):
    """Valid types for parent entity."""

    data_system = "data_system"
    data_source = "data_source"
    data_unit = "data_unit"
    data_product = "data_product"


class ChildEntityType(enum.Enum):
    """Valid types for child entity."""

    data_source = "data_source"
    data_unit = "data_unit"
    data_product = "data_product"
    output = "output"


VALID_LINKS = [
    (ParentEntityType.data_system, ChildEntityType.data_source),
    (ParentEntityType.data_source, ChildEntityType.data_unit),
    (ParentEntityType.data_unit, ChildEntityType.data_product),
    (ParentEntityType.data_product, ChildEntityType.data_product),
    (ParentEntityType.data_product, ChildEntityType.output),
]


def _check_link_types(p: ParentEntityType, c: ChildEntityType) -> None:
    if (p, c) not in VALID_LINKS:
        raise exit_with_output(
            msg=f"Link between {p.value} and {c.value} is not allowed.",
            exit_code=1,
        )


@app.command(name="create")
def link_entities(
    ctx: typer.Context,
    parent_type: ParentEntityType = typer.Argument(..., help="Parent entity type"),
    parent_identifier: str = typer.Argument(..., help="Parent identifier"),
    child_type: ChildEntityType = typer.Argument(..., help="Child entity type"),
    child_identifier: str = typer.Argument(..., help="Child identifier"),
) -> None:
    """Create link between entities."""
    _check_link_types(parent_type, child_type)
    url = _link_url(ctx, parent_type.value, parent_identifier, child_type.value, child_identifier)
    util.post_and_process(ctx, constant.GATEWAY, url)


@app.command(name="delete")
def unlink_entities(
    ctx: typer.Context,
    parent_type: ParentEntityType = typer.Argument(..., help="Parent entity type"),
    parent_identifier: str = typer.Argument(..., help="Parent identifier"),
    child_type: ChildEntityType = typer.Argument(..., help="Child entity type"),
    child_identifier: str = typer.Argument(..., help="Child identifier"),
) -> None:
    """Delete link between entities."""
    _check_link_types(parent_type, child_type)
    url = _link_url(ctx, parent_type.value, parent_identifier, child_type.value, child_identifier)
    util.delete_and_process(ctx, constant.GATEWAY, url)
